/*
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		IP/TCP/UDP checksumming routines
 *
 * Authors:	Jorge Cwik, <jorge@laser.satlink.net>
 *		Arnt Gulbrandsen, <agulbra@nvg.unit.no>
 *		Tom May, <ftom@netcom.com>
 *              Pentium Pro/II routines:
 *              Alexander Kjeldaas <astor@guardian.no>
 *              Finn Arne Gangstad <finnag@guardian.no>
 *		Lots of code moved from tcp.c and ip.c; see those files
 *		for more names.
 *
 * Changes:     Ingo Molnar, converted csum_partial_copy() to 2.1 exception
 *			     handling.
 *		Andi Kleen,  add zeroing on error converted to pure assembler
 *		2002-10-30 Denis Vlasenko
 *			boot-time benchmarking, 3Dnow/MMX+/SSE versions
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */

#include <asm/errno.h>
				
/*
** computes a partial checksum, e.g. for TCP/UDP fragments
**
** unsigned int csum_partial(const unsigned char * buff,
**		int len, unsigned int sum)
*/

#ifdef __KERNEL__
#define K(a...) a
#else
#define K(a...)
#endif

#define SRC(y...) \
9999:	y			;\
	.section __ex_table, "a";\
	.long	9999b, 6001f	;\
	.previous

#define DST(y...) \
9999:	y			;\
	.section __ex_table, "a";\
	.long	9999b, 6002f	;\
	.previous

#define	KERNEL_FPU_BEGIN \
	call	kernel_fpu_begin

#define	KERNEL_FPU_END(r) \
K(	movl	%cr0, r		;)\
K(	orl	$8, r		;)\
K(	movl	r, %cr0		;)

.text

#include "csumcpy_naive.inc"
#include "csumcpy_basic.inc"
#include "csumcpy_ssemmxplus.inc"
#include "csumcpy_sse.inc"

.align 4
.globl csum_partial_copy_generic

csum_partial_copy_generic:
	pushl	%ebx
	pushl	%edi
	pushl	%esi
	pushl	%ebp
	movl	%esp, %ebp

#define STK_DERR 40(%ebp)
#define STK_SERR 36(%ebp)
#define STK_SUM 32(%ebp)
#define STK_LEN 28(%ebp)
#define STK_DST 24(%ebp)
#define STK_SRC 20(%ebp)
#define STK_EIP 16(%ebp)
#define STK_EBX 12(%ebp)
#define STK_EDI 8(%ebp)
#define STK_ESI 4(%ebp)
#define STK_EBP (%ebp)
	
	movl	STK_SRC, %esi	#src
	movl	STK_DST, %edi	#dst	
	movl	STK_LEN, %ecx	#len
	movl	STK_SUM, %eax	#sum

        testl	$3, %edi	# Check dst alignment
        jz	40f

    # not 4-aligned: analyze and align...
	testl	$1, %edi
	jz	30f

    # unaligned start addr
	decl	%ecx
	js	90f 		# sz==0, exit
	movzbl	(%esi), %ebx	# eat one byte...
	movb	%bl, (%edi)
	addl	%ebx, %eax
	adcl	$0, %eax
	roll	$8, %eax	# NB: need to be undone at exit!
	incl	%esi
	incl	%edi
	testl	$2, %edi
	jz	40f
30:
    # 2-aligned, but not 4-aligned
	cmpl	$3, %ecx
	jbe	60f
	movw	(%esi), %bx	# eat 2 bytes
	addw	%bx, %ax
	movw	%bx, (%edi)
	adcl	$0, %eax
	leal	2(%esi), %esi
	leal	2(%edi), %edi
	subl	$2, %ecx
40:
    # edi is 4-aligned now: call block routine
	movl	$csumcpy_basic, %ebx	# 'default', known good for ecx==0 etc
	cmpl	$128, %ecx		# use optimized routine
	jb	50f			#   only for large blocks
	movl	best_csumcpy, %ebx
50:	call	*%ebx
60:
    # handle last 0-3 bytes
	jecxz	80f
	cmpl	$2, %ecx
	jb	70f
SRC(	movw	(%esi), %cx         )
	leal	2(%esi), %esi
DST(	movw	%cx, (%edi)         )
	leal	2(%edi), %edi
	je	75f
	shll	$16, %ecx
70:
SRC(	movb	(%esi), %cl         )
DST(	movb	%cl, (%edi)         )
75:	addl	%ecx, %eax
	adcl	$0, %eax
80:
    # undo csum rotation if dst was unaligned
	testl	$1, STK_DST
	jz	90f
	roll	$8, %eax
90:
	movl	%esp, %ebp
	popl	%ebp
	popl	%esi
	popl	%edi
	popl	%ebx
	ret


.section .fixup, "ax"
6001:	movl	STK_SERR, %ebx		# src_err_ptr	
	movl	$-EFAULT, (%ebx)
	# zero the complete destination (computing the rest is too much work)
	movl	STK_DST, %edi		# dst
	movl	STK_LEN, %ecx		# len
	xorl	%eax, %eax
	cld
	rep; stosb
	jmp	90b
6002:	movl	STK_DERR, %ebx		# dst_err_ptr
	movl	$-EFAULT, (%ebx)
	jmp	90b
.previous
